-- Copyright (c) 2015-present, Facebook, Inc.
-- All rights reserved.
--
-- This source code is licensed under the BSD-style license found in the
-- LICENSE file in the root directory of this source tree. An additional grant
-- of patent rights can be found in the PATENTS file in the same directory.


local class = require 'class'

local Set    = require 'pl.Set'
local List   = require 'pl.List'
local tablex = require 'pl.tablex'

local actions   = require 'babi.actions'
local Task      = require 'babi.Task'
local World     = require 'babi.World'
local Question  = require 'babi.Question'
local Clause    = require 'babi.Clause'
local utilities = require 'babi.utilities'
local Entity    = require 'babi.Entity'

local Counting = class('Counting', 'Task')

function Counting:new_world()
    local world = World()
    world:load((BABI_HOME or '') .. 'tasks/worlds/world_basic.txt')
    return world
end

function Counting:generate_story(world, knowledge, story)
    -- Our story will be 2 statements, 1 question, 5 times
    local allowed_actions = {actions.get, actions.give, actions.teleport, actions.drop}
    local actors = world:get_actors()
    local known_actors = Set()
    local support = {}

    local isIQA = true;
    if isIQA then 
        for i = 1, #actors do
            support[actors[i]] = {}
            local objs         = world:get_objects();
            for j = 1, #objs do
                support[actors[i]][ objs[j] ] = List();
            end
        end
    else
        for i = 1, #actors do
            support[actors[i]] = List()
        end
    end

    local i       = 0
    local loopNum = isIQA and 2 or 8;
    local first_question     = nil;
    local first_random_actor = nil;
    -- make the first person holds multiple objects
    if isIQA then
        local actor_obj_map   = {};
        local minObjNum       = 4;
        local actorCandidates ={};
        for j = 1, #world:get_actors() do
            local currActor = world:get_actors()[j];
            actor_obj_map[currActor] = {};
            for k = 1, #world:get_objects() do
                local currObj = world:get_objects()[k];
                if currObj.is_in == currActor.is_in then
                    actor_obj_map[currActor][ #actor_obj_map[currActor] + 1] = currObj;
                end
            end
            if #actor_obj_map[currActor] >= minObjNum then
                actorCandidates[#actorCandidates + 1] = currActor; 
            end
        end
        first_random_actor       = actorCandidates[math.random(#actorCandidates)];
        local first_random_objs  = actor_obj_map[first_random_actor];
        local first_obj_inds     = torch.randperm(#first_random_objs);
        local validCount         = 0;
        while (validCount < minObjNum) do
            i = i + 1;
            local randomObj = first_random_objs[first_obj_inds[validCount + 1]];
            local clause    = nil;
            local isGet     = nil;
            while not clause do
                isGet = i > 15 or math.random(3) <= 2; 
                if isGet then
                    clause = Clause.sample_valid(world, {true},
                                {first_random_actor},
                                {actions.get}, {randomObj} )
                    if clause then
                        known_actors = known_actors + Set{clause.actor}
                        support[clause.actor][randomObj]:append(clause)
                        validCount = validCount + 1;
                    end
                else
                    local random_actor = world:get_actors()[math.random(#world:get_actors())];
                    while random_actor == first_random_actor do
                        random_actor = world:get_actors()[math.random(#world:get_actors())];
                    end
                    clause = Clause.sample_valid(
                            world, {true}, {random_actor},
                            {actions.teleport}, world:get_locations() )
                end
            end
            clause:perform()
            story:append(clause)
            knowledge:update(clause)
        end
    end
    while not first_question or i - first_question < loopNum do
        -- Find a random action
        local clause
        while not clause do
            local random_action =
                allowed_actions[math.random(#allowed_actions)]
            if class.istype(random_action, 'Teleport') then
                clause = Clause.sample_valid(
                    world, {true}, world:get_actors(),
                    {actions.teleport}, world:get_locations()
                )
            elseif class.istype(random_action, 'Get') then
                local randomObj = world:get_objects()[ math.random(#world:get_objects()) ];

                clause = Clause.sample_valid(
                    world, {true}, world:get_actors(),
                    {actions.get, actions.get}, {randomObj}
                )
                if clause then
                    known_actors = known_actors + Set{clause.actor}
                    if isIQA then
                        support[clause.actor][randomObj]:append(clause)
                    else
                        support[clause.actor]:append(clause)
                    end
                end
            else
                local randomObj = world:get_objects()[ math.random(#world:get_objects()) ];

                clause = Clause.sample_valid(
                    world, {true}, world:get_actors(),
                    {actions.give}, {randomObj}, world:get_actors()
                )
                if clause then
                    known_actors = known_actors + Set{clause.actor}
                    if isIQA then
                        support[clause.actor][randomObj]:append(clause)
                    else
                        support[clause.actor]:append(clause)
                    end
                end
            end
        end
        i = i + 1
        clause:perform()
        story:append(clause)
        knowledge:update(clause)
        if ((not first_question and i > 1 and Set.len(known_actors) > 0) or
                (first_question and (i - first_question) % 2 == 0) ) then
            -- Pick a random actor and ask how many objects he/she is carrying
            local random_actor = tablex.keys(known_actors)[math.random(Set.len(known_actors))];
            if isIQA and not first_question then
                random_actor = first_random_actor;
            end
            local held_objects = List()
            for _, entity in pairs(world:get_objects()) do
                local value = knowledge:current()[entity]:get_value('is_in')
                if value == random_actor then
                    held_objects:append(entity)
                end
            end
            first_question = first_question or i

            if isIQA then
                local askObjNum  = math.random(#held_objects <= 1 and 0 or 1,
                                    #held_objects > 0 and #held_objects - 1 or 0);
                local askObjs    = {};
                local askSupport = {};
                if askObjNum > 0 then
                    askObjs = utilities.choice(held_objects, askObjNum);
                    for j = 1, #askObjs do
                        local sps = support[random_actor][askObjs[j]];
                        for k = 1, #sps do
                            askSupport[#askSupport + 1] = sps[k];
                        end
                    end
                end

                story:append(Question(
                    'count',
                    Clause(world, true, world:god(), actions.set, random_actor,
                           'holding', askObjs),
                    Set(askSupport)
                ))
                story:append(Question(
                    'complementaryFI',
                    Clause(world, true, world:god(), actions.set, askObjs),
                    Set(askSupport)
                ))
            else
                story:append(Question(
                    'count',
                    Clause(world, true, world:god(), actions.set, random_actor,
                           'holding', held_objects),
                    Set(support[random_actor])
                ))
            end
        end
    end
    return story, knowledge
end

return Counting
