-- Copyright (c) 2015-present, Facebook, Inc.
-- All rights reserved.
--
-- This source code is licensed under the BSD-style license found in the
-- LICENSE file in the root directory of this source tree. An additional grant
-- of patent rights can be found in the PATENTS file in the same directory.


local class = require 'class'

local List = require 'pl.List'
local Set = require 'pl.Set'

local actions = require 'babi.actions'
local Task = require 'babi.Task'
local World = require 'babi.World'
local Question = require 'babi.Question'
local Clause = require 'babi.Clause'

local OPPOSITE_DIRECTIONS = {n='s', ne='sw', e='w', se='nw', s='n',
                             sw='ne', w='e', nw='se', u='d', d='u'}

local IsDir = class('IsDir', 'Task')

function IsDir:new_world()
    local world = World()

    -- if generate iqa data or regular qa data
    self.isIQA = true;

    -- random numbers
    self.dir  = ({'n', 'e'})[math.random(2)];

    -- Ask about one of the two relations
    self.ask_dir = ({self.dir, OPPOSITE_DIRECTIONS[self.dir]})[math.random(2)]
    self.ask_location = self.dir == self.ask_dir and 3 or 1

    -- Keep track of which of the two statements supports this
    self.supporting_fact = math.min(self.ask_location, 2)
    --self.supporting_fact = self.swap > 1 and self.supporting_fact % 2 + 1 or self.supporting_fact
    self.switch_supporting_fact = math.random(2) > 1 and true or false;
    --print('switch_supporting_fact: ' .. (self.switch_supporting_fact and 'true' or 'false'))

    local index_loc_in_question = 2;
    local index_loc_in_answer   = self.ask_location;
    local index_loc_amb_not_ans = self.ask_location == 1 and 3 or 1;
    if self.switch_supporting_fact then
         index_loc_in_question, index_loc_in_answer = 
                   index_loc_in_answer, index_loc_in_question;
    end

    -- Pick three random locations
    local options = {'bedroom', 'bathroom', 'kitchen',
                     'office', 'garden', 'hallway'}
    local ambPrefix = {
                    {[1] = 'master'   , [2] = 'guest'},
                    {[1] = 'small'    , [2] = 'large'},
                    {[1] = 'regular'  , [2] = 'special'},
                    {[1] = 'temporary', [2] = 'official'},
                    {[1] = 'small'    , [2] = 'large'},
                    {[1] = 'long'     , [2] = 'short'}};

    local locations = List()
    local locNum    = 6;
    if self.isIQA then
        local _indices = torch.randperm(#options);
        local index_option_regular = _indices[1];
        local index_option_amb     = _indices[2];
        local index_amb_prefix     = torch.randperm(2);

        local _i_amb_prefix = 1;
        local extra_index   = 3;
        while #locations < locNum do
            local option = nil; --options[math.random(#options)]
            if #locations + 1 == index_loc_in_answer then
                option = options[index_option_regular];
            elseif #locations < 3 then
                option = ambPrefix[index_option_amb][index_amb_prefix[_i_amb_prefix]] .. ' ' .. options[index_option_amb];
                _i_amb_prefix = _i_amb_prefix + 1;
            else
                option      = options[_indices[extra_index]];
                extra_index = extra_index + 1;
            end
            if not world.entities[option] then
                local location = world:create_entity(option, {is_location = true})
                locations:append(location)
            end
        end

        self.amb_location_prefix_1    = ambPrefix[index_option_amb][index_amb_prefix[1]];
        self.amb_location_prefix_2    = ambPrefix[index_option_amb][index_amb_prefix[2]];
        self.amb_location_in_question = world:create_entity(options[index_option_amb], {is_location = true})
    else
        while #locations < locNum do
            local option = options[math.random(#options)]
            if not world.entities[option] then
                local location = world:create_entity(option, {is_location = true})
                locations:append(location)
            end
        end 
         self.amb_location_in_question = locations[index_loc_in_question]; 
    end
    self.location_in_answer = locations[index_loc_in_answer];
    if self.switch_supporting_fact then
        self.amb_location_in_question, self.location_in_answer =
                    self.location_in_answer, self.amb_location_in_question;
    end
    self.index_loc_in_question = index_loc_in_question;
    self.index_loc_amb_not_ans = index_loc_amb_not_ans;
    -- Choose the direction in which the locations wlil be ordered
    for i = 1, 3 do
        world:perform_action('set_pos', world:god(), locations[i],
                             self.dir == 'e' and (i - 2) or 0,
                             self.dir == 'n' and (i - 2) or 0)
        if i > 1 then
            world:perform_action('set_dir', world:god(), locations[i - 1],
                                 self.dir, locations[i])
        end
    end
    self.locations = locations
    return world
end

function IsDir:generate_story(world, knowledge, story)
    -- Inform the reader of the two relations

    local _indices       = torch.randperm(4);
    self.supporting_fact = (self.supporting_fact == 1 and
                                _indices[1] or _indices[2]);

    local noiseOpt = true;
    local noiseIdx = nil;
    if noiseOpt then
        local changeIndex = math.random(2) + 3;
        self.locations[changeIndex], self.locations[self.index_loc_amb_not_ans]=
            self.locations[self.index_loc_amb_not_ans], self.locations[changeIndex];
        changeIndex, self.index_loc_amb_not_ans =
            self.index_loc_amb_not_ans, changeIndex;
        noiseOpt = {4, 5, 6};
        if true then -- math.random(2) == 1 then
            local mapIdx  = self.index_loc_in_question; 
            local noanIdx = self.index_loc_amb_not_ans == 4 and 1 or 2;
            noiseOpt[mapIdx], noiseOpt[noanIdx] =
                            noiseOpt[noanIdx], noiseOpt[mapIdx]
        else
            local randIdx = torch.randperm(3);
            local _tmp    = {};
            for i = 1, 3 do
                _tmp[i] = noiseOpt[randIdx[i]];
            end
            noiseOpt = _tmp;
        end
    end

    story[_indices[1]] = Clause(world, true, world:god(), actions.set,
                        self.locations[1], self.dir, self.locations[2]);
    story[_indices[2]] = Clause(world, true, world:god(), actions.set,
                        self.locations[2], self.dir, self.locations[3]);
    if noiseOpt then
        story[_indices[3]] = Clause(world, true, world:god(), actions.set,
            self.locations[noiseOpt[1]], self.dir, self.locations[noiseOpt[2]]);
        story[_indices[4]] = Clause(world, true, world:god(), actions.set,
            self.locations[noiseOpt[2]], self.dir, self.locations[noiseOpt[3]]);
    else
        story[_indices[3]] = Clause(world, true, world:god(), actions.set,
                        self.locations[4], ({'n', 'e'})[math.random(2)],
                            self.locations[self.index_loc_amb_not_ans]);
        story[_indices[4]] = Clause(world, true, world:god(), actions.set,
                        self.locations[4], ({'n', 'e'})[math.random(2)],
                            self.locations[5]);
    end

    -- test
    --print(self.locations[1].name .. ', ' .. self.locations[2].name .. ', ' .. self.locations[3].name)
    --print('dir: ' .. self.dir .. ', ask_dir:' .. self.ask_dir .. ', swap:' .. self.swap ..
    --      ', question:' .. self.locations[2].name .. ', target:'
    --      .. self.ask_location .. ':' .. self.locations[self.ask_location].name );

    -- Ask the question :: 2 is the location in question, self.ask_location is the location for answer,
    -- self.switch_supporting_fact decides whether to swtich them
    story[#story + 1] = Question(
        'eval',
        Clause(world, true, world:god(), actions.set,
            self.amb_location_in_question, self.ask_dir, 
                self.location_in_answer, self.switch_supporting_fact),
        Set{story[self.supporting_fact]}
    )

    if self.isIQA then
        local _amb_object = self.switch_supporting_fact and
                    self.location_in_answer or self.amb_location_in_question;
        -- ask complementary question
        story[#story + 1] = Question(
            'complementary',
            Clause(world, true, world:god(), actions.set, _amb_object,
                    self.amb_location_prefix_1, self.amb_location_prefix_2,
                    self.locations[self.index_loc_in_question]),
            Set{story[self.supporting_fact]}
        )
    end

    return story, knowledge
end

return IsDir
